%     Copyright (C) 2017 Imatest LLC
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program.  If not, see <http://www.gnu.org/licenses/>.


% DESCRIPTION
% 
% This MATLAB script illustrates how to produce a "distortion corrected" image from the radial 
% distortion polynomial coefficients measured in Imatest. 
% 
% Includes examples of how to undo both the distortion described by a direct radial distortion 
% polynomial (as returned by Checkerboard and SFRPlus modules) and the distortion described by a 
% Local Geometric Distortion (LGD) polynomial (as returned by Dot Pattern module).
%
% Some useful variable naming convensions used below:
% 1) '_d' indicates distorted (observed) values, '_u' indicates undistorted (ideal) ones
% 2) Capitalized variables indicate matrices of coordinates used to described pixel locations
%
% See the associated blog post on imatest.com for a more qualitative details about the reasoning 
% behind this approach.



% TO USE THIS EXAMPLE SCRIPT: 
% 
% - Simply run this script in MATLAB from the same folder as the other supplied files
% - As is, this script undoes distortion as measured in SFRPlus module. To have it undo distortion 
%     as measured in Dot Pattern module, comment out/uncomment the appropriate code. Note that the
%     two processes are essentially exactly the same except for which polynomial conversion function
%     is used at the start.
% - This script should work with a stock MATLAB install with no special toolboxes. All functions are
%     built-in basic MATLAB functions except for invert_lgd_poly() and invert_distortion_poly(), 
%     which are supplied.
% - To test distortion correction on your own images, simply change the filepath and 
%     inverseCoeffs/lgdCoeffs variables below to be the path to your image file (relative or 
%     absolute) and the distortion polynomial coefficient values measured by Imatest, respectively.
% - NOTE: You may have to re-order the Imatest-reported coefficients and/or augment with zeros in
%     the appropriate locations so that the polynomials fit MATLAB's expected format. For example: 
%     Imatest reports 3rd order distortion coefficient k = 0.05458 -> MATLAB polynomial of 
%     [0.05458, 0, 1, 0]. See 'help polyval' for description. 
% - NOTE: The examples below use the 3rd order polynomial coefficients, but this all works just as
%     well if you plug in the 5th order ones supplied instead.
% - Try applying the inversion to offices_distortion.jpg to see the correction a real world image
%     instead of a test chart.


% (Start with a clean slate)
clear, clc, close all


% - - UNCOMMENT THIS SEGMENT FOR EXAMPLE USING DISTORTION COEFFICIENTS FROM SFRPLUS/CHECKERBOARD - -
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
filepath = 'sfrplus_distortion.jpg';
inverseCoeffs = [0.2259 0 1 0]; % distortion coefficients reported by SFRPlus module for BOTH
                                % 'sfrplus_distortion.jpg' and 'offices_distortion.jpg'

% Convert the supplied polynomial which transforms r_d -> r_u into a polynomial which transforms 
% r_u -> r_d. Using this, we will transform sampling point locations from an undistorted space to
% those in the distorted image space, to resample the distorted image into an undistorted one.
forwardCoeffs = invert_distortion_poly(inverseCoeffs);
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



% - - - UNCOMMENT THIS SEGMENT FOR EXAMPLE USING LGD COEFFICIENTS FROM DOTPATTERN - - - - - - - - -
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
% filepath = 'dot_distortion.jpg';
% lgdCoeffs = [-18.183, 24.107, -3.653, 0.189]; % lgd coefficients reported by Dot Pattern module
% forwardCoeffs = lgd2forward_poly(lgdCoeffs);
% inverseCoeffs = invert_distortion_poly(forwardCoeffs); % used solely for scaleFactor, below
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



% - - - - - - - - - - - - CODE BELOW HERE IS COMMON TO BOTH EXAMPLES - - - - - - - - - - - - - - -

% Read the image and get its useful dimensions
im_d = double(imread(filepath));
width = size(im_d, 2);
height = size(im_d, 1);
channels = size(im_d, 3);

% Produce a mesh grid of the coordinates of each pixel, relative to the exact center of the image
xs = ((1:width) - (width+1)/2);
ys = ((1:height) - (height+1)/2);
[X, Y] = meshgrid(xs,ys);

% Convert cartesian coordinates of each pixel location to polar, so we can manipulate the radial 
% distances. We define the image-center-to-corner-pixel distance to be the image size, and normalize
% by that so that all pixels have radial distances from 0 to 1. We then find the scale factor which
% will map from radius=1 in the undistorted image to the appropriate point in the distorted image.
[THETA, RHO_d] = cart2pol(X, Y);
normFactor = RHO_d(1, 1); 
scaleFactor = polyval(inverseCoeffs, 1);
RHO_d = RHO_d/normFactor*scaleFactor;

% Apply our distortion-correcting-polynomial to the observed radial distances to get the ideal ones.
RHO_u = polyval(forwardCoeffs, RHO_d);

% Convert back to cartesian coordinates so get the (x,y) distorted sample points in image space
[X_d, Y_d] = pol2cart(THETA, RHO_u*normFactor);

% Re-sample the image at the corrected points using the interp2 function. Apply to each color
% channel independently, since interp2 only works on 2-d data (hence the name). 
im_u = zeros(height,width,channels); % pre-allocate space in memory
for c = 1:channels
   im_u(:,:,c) = interp2(X, Y, im_d(:,:,c), X_d, Y_d);
end


% Display the results, with some helpful grids to verify the correction by eye
figure
image(uint8(im_d))
axis on, grid on, set(gca, 'XTickLabel', '', 'YTickLabel', '', 'GridAlpha', 0.85), axis image
title('Distorted (input) image')

figure
image(uint8(im_u))
axis on, grid on, set(gca, 'XTickLabel', '', 'YTickLabel', '', 'GridAlpha', 0.85), axis image
title('Undistorted (corrected) image')